pub mod audio_processor;
pub mod audio_device;

use cpal::SupportedBufferSize;
use cpal::traits::{DeviceTrait, HostTrait, StreamTrait};
use serde::{Deserialize, Serialize};
use crate::audio_engine::audio_device::get_device_name;

#[derive(Serialize, Deserialize, Debug)]
struct AudioConfig {
    sample_rate: Option<u32>,
    buffer_size: Option<u32>,
    input_device: Option<String>,
    output_device: Option<String>,
}

impl AudioConfig {
    fn load() -> Self {
        let config = std::fs::read_to_string("audio_config.json");
        let default = AudioConfig {
            sample_rate: None,
            buffer_size: None,
            input_device: None,
            output_device: None,
        };
        if config.is_err() {
            println!("读取audio_config.json失败, 将会使用默认配置");
            return default;
        }
        let result = serde_json::from_str(&config.unwrap());
        if result.is_err() {
            println!("解析audio_config.json失败, 将会使用默认配置");
        }
        result.unwrap_or(default)
    }

    fn save(&self) {
        let config = serde_json::to_string(&self).expect("序列化audio_config.json失败");
        std::fs::write("audio_config.json", config).expect("无法写入audio_config.json");
    }
    fn get_sample_rate(&self, default: u32) -> cpal::SampleRate {
        cpal::SampleRate(self.sample_rate.unwrap_or(default))
    }
    fn get_buffer_size(&self, default: cpal::BufferSize) -> cpal::BufferSize {
        let size = self.buffer_size.unwrap_or(0);
        if size == 0 {
            default
        } else {
            cpal::BufferSize::Fixed(size)
        }
    }
    fn get_input_device(&self, default: Option<cpal::Device>) -> Option<cpal::Device> {
        let host = cpal::default_host();
        let input_device = self.input_device.as_ref().map(|name| {
            host.devices().unwrap().find(|device| {
                device.name().unwrap() == *name
            })
        }).flatten();
        input_device.or(default)
    }
    fn get_output_device(&self, default: Option<cpal::Device>) -> Option<cpal::Device> {
        let host = cpal::default_host();
        let output_device = self.output_device.as_ref().map(|name| {
            host.devices().unwrap().find(|device| {
                device.name().unwrap() == *name
            })
        }).flatten();
        output_device.or(default)
    }
}

pub struct AudioEngine {
    host : cpal::Host,
    input_device: Option<cpal::Device>,
    output_device: Option<cpal::Device>,
    sample_rate: cpal::SampleRate,
    buffer_size: cpal::BufferSize,

    input_stream: Option<cpal::Stream>,
    output_stream: Option<cpal::Stream>,
}

impl AudioEngine {
    pub fn new() -> Self {
        println!("Supported hosts:\n  {:?}", cpal::ALL_HOSTS);

        let default_host = cpal::host_from_id(cpal::HostId::Asio).unwrap_or(cpal::default_host());
        let default_output_device = default_host.default_output_device();
        let output_config = default_output_device.as_ref().unwrap().default_output_config().unwrap();
        let sample_rate = output_config.sample_rate();

        let audio_config = AudioConfig::load();

        let out = AudioEngine {
            host: default_host,
            input_device: audio_config.get_input_device(None),
            output_device: audio_config.get_output_device(default_output_device),
            sample_rate: audio_config.get_sample_rate(sample_rate.0),
            buffer_size: audio_config.get_buffer_size(cpal::BufferSize::Default),
            input_stream: None,
            output_stream: None,
        };
        out.save_config();

        let input_device_name = get_device_name(out.input_device.as_ref());
        let output_device_name = get_device_name(out.output_device.as_ref());
        println!("Using API: {}", out.host.id().name());
        println!("Input device: {}", input_device_name);
        println!("Output device: {}", output_device_name);
        out
    }

    pub fn save_config(&self) {
        let audio_config = AudioConfig {
            sample_rate: Some(self.sample_rate.0),
            buffer_size: Some(match self.buffer_size {
                cpal::BufferSize::Default => { 0 }
                cpal::BufferSize::Fixed(size) => { size }
            }),
            input_device: Some(get_device_name(self.input_device.as_ref())),
            output_device: Some(get_device_name(self.output_device.as_ref())),
        };
        audio_config.save();
    }

    pub fn open_stream(&mut self) {
        self.try_open_input_stream();
        self.try_open_output_stream();
    }

    fn try_open_output_stream(&mut self) {
        if self.output_device.is_none() {
            return;
        }
        let device = self.output_device.as_ref().unwrap();
        let config = self.output_stream_config();
        let output_stream = device.build_output_stream(
            &config,
            |data, callback| process_output(data, callback),
            move |err| eprintln!("an error occurred on stream: {}", err),
            None
        );
        if output_stream.is_err() {
            eprintln!("failed to open output stream: {}", output_stream.err().unwrap());
            return;
        }

        self.output_stream = Some(output_stream.unwrap());
        let play_result = self.output_stream.as_mut().unwrap().play();
        if play_result.is_err() {
            eprintln!("failed to play output stream: {}", play_result.err().unwrap());
        }
    }

    fn try_open_input_stream(&mut self) {
        if self.input_device.is_none() {
            return;
        }
        let device = self.input_device.as_ref().unwrap();
        let input_stream = device.build_input_stream(
            &self.input_stream_config(),
            |data, callback| process_input(data, callback),
            move |err| eprintln!("an error occurred on stream: {}", err),
            None
        ).unwrap();
        input_stream.play().unwrap();
        self.input_stream = Some(input_stream);
    }

    fn output_stream_config(&self) -> cpal::StreamConfig {
        cpal::StreamConfig {
            channels: 2,
            buffer_size: self.buffer_size,
            sample_rate: self.sample_rate,
        }
    }

    fn input_stream_config(&self) -> cpal::StreamConfig {
        cpal::StreamConfig {
            channels: 2,
            buffer_size: self.buffer_size,
            sample_rate: self.sample_rate,
        }
    }
}

fn process_output(data: &mut [f32], callback_info: &cpal::OutputCallbackInfo) {
    for sample in data.iter_mut() {
        *sample = 0.0;
    }
}

fn process_input(data: &[f32], callback_info: &cpal::InputCallbackInfo) {

}
