extern crate vst2;

mod host;

use std::sync::{Arc, Mutex};
use vst2::buffer::AudioBuffer;
use vst2::host::*;
use vst2::plugin::*;
use vst2::event::Event;
use crate::host::HostEvent;
use host::HostInterface;

impl HostEvent {
	fn to_vst2_event(&self) -> Event {
		match self {
			HostEvent::Midi { data, delta_frames, live, note_length, note_offset, detune, note_off_velocity } => {
				Event::Midi {
					data: *data,
					delta_frames: *delta_frames,
					live: *live,
					note_length: note_length.map(|x| *x),
					note_offset: note_offset.map(|x| *x),
					detune: *detune,
					note_off_velocity: *note_off_velocity,
				}
			},
			HostEvent::SysEx { payload, delta_frames } => {
				Event::SysEx {
					payload,
					delta_frames: *delta_frames,
				}
			},
		}
	}
}

struct Vst2Host { on_automate: Option<Box<dyn Fn(i32, f32) + Send>> }

impl Host for Vst2Host {
	fn automate(&self, index: i32, value: f32) {
		self.on_automate.as_ref().map(|f| f(index, value));
	}
}

struct InternalVst2Host {
	host: Arc<Mutex<Vst2Host>>,
	loader: PluginLoader<Vst2Host>,
	instance: PluginInstance,
	window: Option<Window>,
}

fn on_window_update(window: &mut Window, elapsed: std::time::Duration) {

}

impl InternalVst2Host {
	fn new(path: &std::path::Path) -> Self {
		let host = Arc::new(Mutex::new(Vst2Host{
			on_automate: None,
		}));

		let mut loader = PluginLoader::load(path, host.clone()).unwrap();
		let mut instance = loader.instance().unwrap();
		let editor = instance.get_editor();

		let mut window : Option<Window> = None;
		if editor.is_some() {
			let w = Window::new(instance.get_info().name.as_mut(), on_window_update);
			let editor_size = editor.unwrap().size();
			window = Some(w);
		}

		InternalVst2Host {
			host,
			loader,
			instance,
			window,
		}
	}

	fn on_automate(&self, index: i32, value: f32) {
		println!("Automate: {} {}", index, value);
	}
}

impl HostInterface for InternalVst2Host {
	fn get_host_name(&self) -> String {
		self.instance.get_info().name
	}

	fn get_host_version(&self) -> String {
		self.instance.get_info().version.to_string()
	}

	fn get_host_vendor(&self) -> String {
		self.instance.get_info().vendor
	}

	fn set_parameter(&mut self, index: i32, value: f32) {
		self.instance.set_parameter(index, value);
	}

	fn get_parameter(&self, index: i32) -> f32 {
		self.instance.get_parameter(index)
	}

	fn process_f32(&mut self, input: Vec<&mut [f32]>, output: Vec<&mut [f32]>) {
		let buffer = AudioBuffer::new(input, output);
		self.instance.process(buffer);
	}

	fn process_f64(&mut self, input: Vec<&mut [f64]>, output: Vec<&mut [f64]>) {
		let buffer = AudioBuffer::new(input, output);
		self.instance.process_f64(buffer);
	}

	fn process_events(&mut self, events: Vec<HostEvent>) {
		let events = events.iter().map(|event| event.to_vst2_event()).collect();
		self.instance.process_events(events);
	}

	fn open_editor(&mut self) {
		let editor = self.instance.get_editor();
		editor.unwrap().open();
	}

	fn close_editor(&mut self) {
		todo!()
	}

	fn is_editor_open(&self) -> bool {
		todo!()
	}

	fn has_editor(&mut self) -> bool {
		self.instance.get_editor().is_some()
	}
}

fn main() {
	// 获取参数
	let args: Vec<String> = std::env::args().collect();
	// 参数1是插件路径
	let path = std::path::Path::new(&args[1]);
	let internal_host = InternalVst2Host::new(path);

	let mut shmem = Sheme;
}
